/*
 * SPMPolynomial.cpp
 *
 * Author: Daniel Hernandez <http://daniel.degu.cl>
 */

#include <vector>
#include <string>
#include <iostream>
#include <stdexcept>

namespace SPM
{  
  /**
   * Enumeration for polynomial node types.
   *
   * A polynomial is essentially an ordered tree which nodes belong to
   * one of the followong types:
   *
   * - `Constant`: a natural number, e.g., 0, 1, 2, and 3.
   * - `Variable`: a variable, e.g., x.
   * - `Sum`: a sum between several polynomials, e.g., (+ 1 x y).
   * - `Product`: a product between several polynomials, e.g., (* 1 x 3).
   * - `Difference`: a difference between two polynomials, e.g., (- 1 x).
   *
   * In the examples above we used Lisp notation.  For instance, the
   * polynomial x + 2y is denoted as (+ x (* 2 y)).  The first element
   * of the list represent the type of the node whereas the rest
   * represent its children.  Nodes 2 and y are of types Constant and
   * Variable, and have no children.  Symbols +, *, and - are used to
   * denote nodes of type `Sum`, `Product`, and `Difference`.
   */
  enum class Operation
    {
      Constant,
      Variable,
      Sum,
      Product,
      Difference
    };

  /**
   * Class for polynomials.
   *
   * As is explained in the description of class `Operation`, a
   * polynomial is essentially an ordered tree.  The template
   * parameter <VarT> allows to create differnt types of polynomials
   * (e.g., polynomials which variables are numeric identifiers or
   * string codidifed URLs).
   *
   * Note: We assume that `VarT` is not `unsigned int`.
   */
  template <class VarT = std::string>
  class Polynomial
  {
  public:
    /**
     * Creates a zero polynomial, i.e., polynomial 0.
     */
    Polynomial():
      _operation(Operation::Constant),
      _constant(0) { }
    /**
     * Creates a constant polynomial (e.g., 0, 1, and 3).
     *
     * @param constant The value of the polynomial.
     */
    Polynomial(unsigned int constant):
      _operation(Operation::Constant),
      _constant(constant) { }
    /**
     * Creates a variable polynomialpolynomial (e.g., `x`).
     *
     * @param variable The varibale of the polynomial.
     */
    Polynomial(VarT variable):
      _operation(Operation::Variable),
      _variable(variable) { }
    /**
     * Creates a polynomials that operates over a list of given
     * polynomial operands.
     *
     * @param operation The operation.
     * @param operands A vector with the operands.
     */
    Polynomial(const Operation operation,
               std::vector<Polynomial<VarT>> operands):
      _operation(operation),
      _operands(operands) { }
    
    /**
     * Writes a polynomial to an output stream.
     */
    template <class VarType>
    friend std::ostream& operator<<(std::ostream& os,
                                    const Polynomial<VarType>& polynomial);

    Polynomial<VarT>
    operator+(const Polynomial<VarT> polynomial)
    {
      std::vector<Polynomial<VarT>> operands = {*this, polynomial};
      return Polynomial<VarT>(Operation::Sum, operands);
    }
    Polynomial<VarT>
    operator*(const Polynomial<VarT> polynomial)
    {
      std::vector<Polynomial<VarT>> operands = {*this, polynomial};
      return Polynomial<VarT>(Operation::Product, operands);
    }
    Polynomial<VarT>
    operator-(const Polynomial<VarT> polynomial)
    {
      std::vector<Polynomial<VarT>> operands = {*this, polynomial};
      return Polynomial<VarT>(Operation::Difference, operands);
    }
  private:
    Operation _operation;
    unsigned int _constant;
    VarT _variable;
    std::vector<Polynomial<VarT>> _operands;
  };

  template <class VarT>
  std::ostream& operator<<(std::ostream& os,
                           const Polynomial<VarT>& polynomial)
  {
    switch(polynomial._operation) {
    case Operation::Constant:
      os << polynomial._constant;
      break;
    case Operation::Variable:
      os << "( variable " << polynomial._variable << " )";
      break;
    default:
      switch(polynomial._operation) {
      case Operation::Product:
        os << "( * ";
        break;
      case Operation::Sum:
        os << "( + ";
        break;
      case Operation::Difference:
        os << "( - ";
        break;
      default:
        throw std::invalid_argument("Unsupported operator type");
      }
      for (auto operand: polynomial._operands)
        os << operand << " ";
      os << ")";
    }
    return os;
  }
}
