/*
 * Author: Daniel Hernandez <http://daniel.degu.cl>
 */

#include "rewritter.hpp"
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <random>
#include <regex>
#include <algorithm>
#include <sys/time.h>
#include <string>
#include <boost/algorithm/string.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/regex.hpp>

namespace qi = boost::spirit::qi;

/**
 * Get the variables occurring in a BGP.
 *
 * @param bgp A BGP.
 * @return The variables occurring in the BGP.
 */
std::vector<std::string>
get_variables(std::string bgp)
{
  std::vector<std::string> variables;
  std::vector<std::string> tokens;
  boost::split(tokens, bgp, boost::is_any_of(" ."));

  for (auto token: tokens)
    {
      boost::trim(token);
      if (token.substr(0,1) == "?")
        variables.push_back(token);
    }

  std::sort(variables.begin(), variables.end());
  auto last = std::unique(variables.begin(), variables.end());
  variables.erase(last, variables.end());

  return variables;
}

/**
 * Separate a triple pattern in components.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @return A vector with the subject, predicate, and object compoents.
 */
std::vector<std::string>
triple_pattern_components(std::string triple_pattern)
{
  std::vector<std::string> triple_components;

  qi::parse(triple_pattern.begin(), triple_pattern.end(),
            qi::as_string[ *(qi::char_ - " ") ] >> " " >>
            qi::as_string[ *(qi::char_ - " ") ] >> " " >>
            qi::as_string[ *(qi::char_) ],
            triple_components);

  return triple_components;
}

/**
 * Returns the pattern that reifies a triple pattern using the named graph scheme.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
named_graphs_reify(std::string triple_pattern,
                   std::string entity)
{
  return "GRAPH " + entity + " { " + triple_pattern + " } .\n";
}

/**
 * Returns the pattern that reifies a triple pattern using the
 * standard reification scheme.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
standard_reify(std::string triple_pattern,
               std::string entity)
{
  std::vector<std::string> triple_components =
    triple_pattern_components(triple_pattern);

  return entity +
    " <http://www.w3.org/1999/02/22-rdf-syntax-ns#subject> " + triple_components[0] + " ;"
    " <http://www.w3.org/1999/02/22-rdf-syntax-ns#predicate> " + triple_components[1] + " ;"
    " <http://www.w3.org/1999/02/22-rdf-syntax-ns#object> " + triple_components[2] + " .\n";
}

/**
 * Returns the pattern that reifies a triple pattern using the
 * Wikidata reification scheme.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
wikidata_reify(std::string triple_pattern,
               std::string entity)
{
  std::vector<std::string> triple_components =
    triple_pattern_components(triple_pattern);

  std::string predicate_p =
    boost::replace_first_copy(triple_components[1], "/direct/", "/");
  std::string predicate_ps =
    boost::replace_first_copy(triple_components[1], "/direct/", "/statement/");

  return
    triple_components[0] + " " + predicate_p + " " + entity + " . " +
    entity + " " + predicate_ps + " " + triple_components[2] + " .\n";
}

/**
 * Returns the pattern that reifies a triple pattern using a variant
 * of the Wikidata reification scheme.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
wikidata_variant_reify(std::string triple_pattern,
                       std::string entity)
{
  std::vector<std::string> triple_components =
    triple_pattern_components(triple_pattern);

  std::string predicate_p =
    boost::replace_first_copy(triple_components[1],
                              "http://", "http://example.org/p/");
  std::string predicate_ps =
    boost::replace_first_copy(triple_components[1],
                              "http://", "http://example.org/ps/");

  return
    triple_components[0] + " " + predicate_p + " " + entity + " . " +
    entity + " " + predicate_ps + " " + triple_components[2] + " .\n";

}

/**
 * Generate a BIND statement for a set of variables and an URL.
 *
 * When evaluated, the bind statement generates an URL from the values
 * of a given set of variables.  The value is bind to the given
 * variable name, that is a provenance variable.
 *
 * @param variables A set of variables.
 * @param as_variable The variable name for the value identifying the value of the
 *   variables.
 * @return The bind statement.
 */
std::string
bind_statement(std::vector<std::string> variables,
               std::string as_variable)
{
  std::string bind_statement_string =
    "BIND (IRI(CONCAT(\"http://example.org/" + as_variable + "?\",";

  size_t i = 1;
  for (auto variable: variables)
    {
      bind_statement_string +=
        "\"" + variable.substr(1,variable.size()) + "=\",COALESCE(" +
        "ENCODE_FOR_URI(xsd:string(" + variable + ")),\"unbound\")";
      if (i < variables.size())
        bind_statement_string += ",\"&\",";
      i++;
    }
  bind_statement_string += ")) AS " + as_variable + ")\n";

  return bind_statement_string;
}

/**
 * Reifies the triples in a BGP pattern.
 *
 * @param query A BGP.
 * @param prefix The prefix used for the variable for the entity.
 * @param reify The function to reify each triple pattern in the BGP.
 * @return The rewritten query and its variables.
 */
RewrittenQuery
bgp_rewrite(std::string query,
            std::string prefix,
            std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rewritten_query; // The output for this function.

  // Generate a vector with all triple patterns.
  std::vector<std::string> triple_patterns;
  qi::parse(query.begin(), query.end(),
            qi::as_string[ *(qi::char_ - " .") ] % (qi::lit(" .")),
            triple_patterns);

  // Reify each triple pattern.  For each triple pattern, the reified
  // pattern and the provenance variable are recorded in
  // `rewritten_query`.
  int i = 1;
  for (auto triple_pattern: triple_patterns) {
    boost::trim(triple_pattern);
    if (triple_pattern != "") {
      std::string provenance_variable = prefix + "sum_product_" + std::to_string(i) + "_statement";
      rewritten_query.query += reify(triple_pattern, provenance_variable);
      rewritten_query.provenance_variables.push_back(provenance_variable);
      i++;
    }
  }

  // Record the variables in the basic graph pattern.
  rewritten_query.solution_variables = get_variables(query);

  // Add provenance variables for sum and product.
  rewritten_query.query +=
    bind_statement(rewritten_query.solution_variables, prefix + "sum") +
    bind_statement(rewritten_query.solution_variables, prefix + "sum_product");
  rewritten_query.provenance_variables.push_back(prefix + "sum");
  rewritten_query.provenance_variables.push_back(prefix + "sum_product");

  return rewritten_query;
}

/**
 */
RewrittenQuery
basic_rewrite(std::string query,
              std::string prefix,
              std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rewritten_query; // The output for this function.

  std::vector<std::string> query_components;
  qi::parse(query.begin(), query.end(),
            "SELECT " >>
            qi::as_string[*(qi::char_ - " WHERE ")] >> " WHERE { " >>
            qi::as_string[*(qi::char_ - " }")] >> " }",
            query_components);

  RewrittenQuery rewritten_bgp =
    bgp_rewrite(query_components[1], prefix + "sum_", reify);

  rewritten_query.solution_variables = get_variables(query_components[0]);
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              rewritten_bgp.provenance_variables.begin(),
                                              rewritten_bgp.provenance_variables.end());
  rewritten_query.provenance_variables.push_back(prefix + "sum");

  rewritten_query.query =
    "SELECT\n" +
    rewritten_query.solution_variables_str() +
    rewritten_query.provenance_variables_str() +
    "WHERE {\n" +
    rewritten_bgp.query +
    bind_statement(get_variables(query_components[0]), prefix + "sum") +
    "}";

  return rewritten_query;
}

/**
 * Rewrites queries with UNION as in the Wikidata benchmark.
 */
RewrittenQuery
union_rewrite(std::string query,
              std::string prefix,
              std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rewritten_query; // The output for this function.

  // A vector with the text before and after WHERE
  std::vector<std::string> query_components;
  qi::parse(query.begin(), query.end(),
            "SELECT " >>
            qi::as_string[*(qi::char_ - " WHERE ")] >> " WHERE { { " >>
            qi::as_string[*(qi::char_ - " } UNION { ")] >> " } UNION { " >>
            qi::as_string[*(qi::char_ - " } } }")] >> " } } }",
            query_components);
  query_components[2] += " }";

  // The inner patterns.
  RewrittenQuery pattern1 =
    rewrite(query_components[1], prefix + "sum_osum_1_", reify);
  RewrittenQuery pattern2 =
    rewrite(query_components[2], prefix + "sum_osum_2_", reify);

  // The solution variables.
  rewritten_query.solution_variables = get_variables(query_components[0]);

  // The provenance variables.
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern1.provenance_variables.begin(),
                                              pattern1.provenance_variables.end());
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern2.provenance_variables.begin(),
                                              pattern2.provenance_variables.end());
  rewritten_query.provenance_variables.push_back(prefix + "sum");
  rewritten_query.provenance_variables.push_back(prefix + "sum_osum");

  // Variables for the subpattern are required for the `sum_osum`
  // provenance variable.
  std::vector<std::string> sum_osum_variables;
  sum_osum_variables.insert(sum_osum_variables.end(),
                            pattern1.solution_variables.begin(),
                            pattern1.solution_variables.end());
  sum_osum_variables.insert(sum_osum_variables.end(),
                            pattern2.solution_variables.begin(),
                            pattern2.solution_variables.end());
  std::sort(sum_osum_variables.begin(), sum_osum_variables.end());
  auto last = std::unique(sum_osum_variables.begin(), sum_osum_variables.end());
  sum_osum_variables.erase(last, sum_osum_variables.end());

  // The rewritten query.
  rewritten_query.query =
    "SELECT\n" +
    rewritten_query.solution_variables_str() +
    rewritten_query.provenance_variables_str() +
    "WHERE {\n{\n" +
    pattern1.query +
    "\n}\nUNION\n{\n" +
    pattern2.query + "\n}\n" +
    bind_statement(rewritten_query.solution_variables, prefix + "sum") +
    bind_statement(sum_osum_variables, prefix + "sum_osum") +
    "}";

  return rewritten_query;
}

RewrittenQuery
optional_rewrite (std::string query,
                  std::string prefix,
                  std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rewritten_query; // The output for this function.

  std::vector<std::string> query_components;
  qi::parse(query.begin(), query.end(),
            "SELECT " >>
            qi::as_string[*(qi::char_ - " WHERE ")] >> " WHERE { " >>
            qi::as_string[*(qi::char_ - " OPTIONAL ")] >> " OPTIONAL { " >>
            qi::as_string[*(qi::char_ - " } . }")] >> " } . }",
            query_components);

  // The subpatterns pattern.
  RewrittenQuery pattern1 =
    bgp_rewrite(query_components[1] + query_components[2],
                prefix + "sum_osum_1_", reify);
  RewrittenQuery pattern2 =
    bgp_rewrite(query_components[1], prefix + "sum_osum_2_difference_1_", reify);
  RewrittenQuery pattern3 =
    bgp_rewrite(query_components[2], prefix + "sum_osum_2_difference_2_sum_", reify);

  // The solution variables.
  rewritten_query.solution_variables = get_variables(query_components[0]);

  // The provenance variables.
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern1.provenance_variables.begin(),
                                              pattern1.provenance_variables.end());
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern2.provenance_variables.begin(),
                                              pattern2.provenance_variables.end());
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern3.provenance_variables.begin(),
                                              pattern3.provenance_variables.end());
  rewritten_query.provenance_variables.push_back(prefix + "sum");
  rewritten_query.provenance_variables.push_back(prefix + "sum_osum");
  rewritten_query.provenance_variables.push_back(prefix + "sum_osum_2_difference");
  rewritten_query.provenance_variables.push_back(prefix + "sum_osum_2_difference_2_sum");

  // The rewritten query.
  rewritten_query.query =
    "SELECT " +
    rewritten_query.solution_variables_str() +
    rewritten_query.provenance_variables_str() +
    "WHERE {\n" +
    "{\n" +
    pattern1.query +
    "}\n" +
    "UNION\n"
    "{\n" +
    pattern2.query +
    "OPTIONAL\n{\n" +
    pattern3.query +
    "}\n" +
    bind_statement(pattern2.solution_variables, prefix + "sum_osum_2_difference") +
    bind_statement(pattern2.solution_variables, prefix + "sum_osum_2_difference_2_sum") +
    "}\n" +
    bind_statement(rewritten_query.solution_variables, prefix + "sum") +
    bind_statement(pattern1.solution_variables, prefix + "sum_osum") +
    "}";

  return rewritten_query;
}

/**
 * Rewrites queries with MINUS as in the Wikidata benchmark.
 *
 * These queries have the following format:
 *
 * ```
 * $prefixes SELECT $vars WHERE { $bgp MINUS { $bpp } }
 * ```
 *
 * And result in a query of the form:
 *
 * ```
 * $prefixes
 * SELECT
 *   $vars
 *   ?prov_sum
 *   ?prov_sum_minus_1_sum
 *   ?prov_sum_minus_1_sum_product
 *   $reify_vars_for_bgp1
 *   ?prov_sum_minus_2_sum
 *   ?prov_sum_minus_2_sum_product
 *   $reify_vars_for_bgp2
 * WHERE {
 *   $reified_bgp1
 *   OPTIONAL
 *   { $reified_bpp2 }
 *   BIND (b(get_variables($bgp_1)) AS ?prov_sum_minus_1_sum_product)
 *   BIND (b(get_variables($bgp_2)) AS ?prov_sum_minus_2_sum_product)
 *   BIND (b(get_variables($bgp_1)) AS ?prov_sum_minus_1_sum)
 *   BIND (b(get_variables($bgp_1)) AS ?prov_sum_minus_2_sum)
 *   BIND (b(get_variables($bgp_1)) AS ?prov_sum)
 * }
 * ```
 */
RewrittenQuery
minus_rewrite (
  std::string query,
  std::string prefix,
  std::string (*reify)(std::string, std::string)
)
{
  RewrittenQuery rewritten_query; // The output for this function.

  // A vector with the query components.
  std::vector<std::string> query_components;
  qi::parse(query.begin(), query.end(),
            "SELECT " >>
            qi::as_string[*(qi::char_ - " WHERE ")] >> " WHERE { " >>
            qi::as_string[*(qi::char_ - " MINUS ")] >> " MINUS { " >>
            qi::as_string[*(qi::char_ - " } }")] >> " } }",
            query_components);

  // The minuend and subtrahend pattern.
  RewrittenQuery minuend =
    bgp_rewrite(query_components[1], prefix + "sum_difference_1_", reify);
  RewrittenQuery subtrahend =
    bgp_rewrite(query_components[2], prefix + "sum_difference_2_sum_", reify);

  // The solution variables.
  rewritten_query.solution_variables = get_variables(query_components[0]);

  // The provenance variables.
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              minuend.provenance_variables.begin(),
                                              minuend.provenance_variables.end());
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              subtrahend.provenance_variables.begin(),
                                              subtrahend.provenance_variables.end());
  rewritten_query.provenance_variables.push_back(prefix + "sum");
  rewritten_query.provenance_variables.push_back(prefix + "sum_difference");
  rewritten_query.provenance_variables.push_back(prefix + "sum_difference_2_sum");

  // The rewritten query.
  rewritten_query.query =
    "SELECT " +
    rewritten_query.solution_variables_str() +
    rewritten_query.provenance_variables_str() +
    "WHERE {\n" +
    minuend.query +
    "OPTIONAL {\n" +
    subtrahend.query +
    "}\n" +
    bind_statement(rewritten_query.solution_variables, prefix + "sum") +
    bind_statement(minuend.solution_variables, prefix + "sum_difference") +
    bind_statement(minuend.solution_variables, prefix + "sum_difference_2_sum") +
    "}";

  return rewritten_query;
}

RewrittenQuery
filter_rewrite(std::string query,
               std::string prefix,
               std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rewritten_query; // The output for this function.

  // A vector with the query components.
  std::vector<std::string> query_components;
  qi::parse(query.begin(), query.end(),
            "SELECT " >>
            qi::as_string[*(qi::char_ - " WHERE ")] >> " WHERE { " >>
            qi::as_string[*(qi::char_ - " FILTER ")] >> " FILTER " >>
            qi::as_string[*(qi::char_ - " }")] >> " }",
            query_components);

  // The subpattern.
  RewrittenQuery pattern =
    bgp_rewrite(query_components[1], prefix + "sum_", reify);

  // The solution variables.
  rewritten_query.solution_variables = get_variables(query_components[0]);

  // The provenance variables.
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern.provenance_variables.begin(),
                                              pattern.provenance_variables.end());
  rewritten_query.provenance_variables.push_back(prefix + "sum");

  // The rewritten query.
  rewritten_query.query =
    "SELECT " +
    rewritten_query.solution_variables_str() +
    rewritten_query.provenance_variables_str() +
    "WHERE {\n" +
    pattern.query +
    "FILTER " +
    query_components[2] + "\n" +
    bind_statement(rewritten_query.solution_variables, prefix + "sum") +
    "}";

  return rewritten_query;
}

RewrittenQuery
filter_group_by_rewrite(std::string query,
                        std::string prefix,
                        std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rewritten_query; // The output for this function.

  // A vector with the query components.
  std::vector<std::string> query_components;
  qi::parse(query.begin(), query.end(),
            "SELECT " >>
            qi::as_string[*(qi::char_ - " WHERE ")] >> " WHERE { " >>
            qi::as_string[*(qi::char_ - " FILTER ")] >> " FILTER " >>
            qi::as_string[*(qi::char_ - " } GROUP BY ")] >> " } GROUP BY " >>
            qi::as_string[*(qi::char_)],
            query_components);

  // The subpattern.
  RewrittenQuery pattern =
    bgp_rewrite(query_components[1], prefix + "aggregate_", reify);

  // The grouping variables.
  std::vector<std::string> solution_variables;
  qi::parse(query_components[0].begin(), query_components[0].end(),
            qi::as_string[*(qi::char_ - "(")] >>
            qi::as_string[*(qi::char_ - " as ")] % (qi::lit(" as ")),
            solution_variables);
  rewritten_query.solution_variables = get_variables(solution_variables[0]);
  solution_variables.erase(solution_variables.begin(),
                           solution_variables.begin() + 2);
  boost::regex re("\\).*$");
  for(auto var: solution_variables)
     rewritten_query.solution_variables.push_back(boost::regex_replace(var, re, ""));

  // The provenance variables.
  rewritten_query.provenance_variables.insert(rewritten_query.provenance_variables.end(),
                                              pattern.provenance_variables.begin(),
                                              pattern.provenance_variables.end());
  rewritten_query.provenance_variables.push_back(prefix + "aggregate");

  std::string bind_sum_variable;
  if (query_components[3] == "(1 AS ?x)")
    bind_sum_variable = "BIND (<http://example.org/?prov_sum> AS " + prefix + "sum)\n";
  else
    bind_sum_variable = bind_statement(rewritten_query.solution_variables, prefix + "sum");

  // The rewritten query.
  rewritten_query.query =
    "SELECT \n" +
    rewritten_query.solution_variables_str() +
    rewritten_query.provenance_variables_str() +
    "WHERE {\n\{\n" +
    query +
    "\n} .\n\{\nSELECT " +
    query_components[3] +
    "\nWHERE {\n" +
    pattern.query +
    "FILTER " +
    query_components[2] + "\n" +
    bind_sum_variable +
    "}\n}\n}";

  return rewritten_query;
}

/**
 * Rewrites a query to compute provenance.
 *
 * @param query The query to rewrite.
 * @param reify A function that reifies triple patterns.
 * @return The reified query.
 */
RewrittenQuery
rewrite(std::string query,
        std::string prefix,
        std::string (*reify)(std::string, std::string))
{
  // We do not avoid DISTINC results because duplicate answers
  // have different provenance.
  boost::replace_first(query, "DISTINCT", "");
  boost::trim(query);

  // We remove duplicate spaces to simplify the parsing.
  boost::regex re("\\s+");
  query = boost::regex_replace(query, re, " ");

  // Case: queries with UNION in the Wikidata benchmark.
  if (query.find("UNION") != std::string::npos)
    return union_rewrite(query, prefix, reify);

  // Case: queries with OPTIONAL in the Wikidata benchmark.
  if (query.find("OPTIONAL") != std::string::npos)
    return optional_rewrite(query, prefix, reify);

  // Case: queries with MINUS in the Wikidata benchmark.
  if (query.find("MINUS") != std::string::npos)
    return minus_rewrite(query, prefix, reify);

  // Case: TPC-H queries.
  // All include FILTER, some include GROUP BY.
  if (query.find("FILTER") != std::string::npos)
    {
      if (query.find("GROUP BY") != std::string::npos) {
        return filter_group_by_rewrite(query, prefix, reify);
      }
      else {
        return filter_rewrite(query, prefix, reify);
      }
    }

  // Case: Basic query in WatDiv and Wikidata benchmark
  if (query.find("SELECT") != std::string::npos)
    return basic_rewrite(query, prefix, reify);

  // Case: A BGP pattern.
  return bgp_rewrite(query, prefix, reify);
}

/**
 * Rewrites a query to compute provenance.
 *
 * @param query The query to rewrite.
 * @param reify A function that reifies triple patterns.
 * @return The reified query.
 */
RewrittenQuery
rewrite_with_prefix(std::string query,
		    std::string prefix,
		    std::string (*reify)(std::string, std::string))
{
  RewrittenQuery rq = rewrite(query, prefix, reify);

  return RewrittenQuery("PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>\n" + rq.query,
			rq.solution_variables,
			rq.provenance_variables);
}
