/*
 * Author: Daniel Hernandez <http://daniel.degu.cl>
 */

#include <string>
#include <vector>

/**
 * Returns the pattern that reifies a triple pattern using the named graph scheme.
 * 
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
named_graphs_reify(std::string triple_pattern,
                   std::string entity);

/**
 * Returns the pattern that reifies a triple pattern using the
 * standard reification scheme.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
standard_reify(std::string triple_pattern,
               std::string entity);

/**
 * Returns the pattern that reifies a triple pattern using the
 * Wikidata reification scheme.
 *
 * The Wikidata reification scheme uses Wikidata prefixes to define
 * two predicates for each predicate.  For example, the URL
 * `<http://www.wikidata.org/prop/direct/P17>` by the URLs
 * `<http://www.wikidata.org/prop/P17>` and
 * `<http://www.wikidata.org/prop/statement/P17>`.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
wikidata_reify(std::string triple_pattern,
               std::string entity);

/**
 * Returns the pattern that reifies a triple pattern using a variant
 * of the Wikidata reification scheme.
 *
 * Unlike the Wikidata reification scheme, this scheme is generic,
 * i.e., can be used for any URL.  For example, the URL `<http://x>`
 * by the URLs `<http://example.org/p/x>` and
 * `<http://example.org/ps/x>`.
 *
 * @param triple_pattern A string that encodes the triple pattern.
 * @param entity A string that encodes the entitity that reifies the triple pattern.
 * @return A string that codifies the reified triple pattern.
 */
std::string
wikidata_variant_reify(std::string triple_pattern,
                       std::string entity);

/**
 * A type for the result of a query rewriting.  Its fields are `query`
 * -- the rewritten query --, `solution_variables` -- the variables
 * that are part of solution --, and `provenance_variables` -- the
 * variables that encode the provenance.
 */
class RewrittenQuery {
protected:
  std::string vars_str(std::vector<std::string> variables) {
    std::string vars = "";
    for(auto variable: variables)
      vars += variable + "\n";

    return vars;
  }
public:
  std::string query;
  std::vector<std::string> solution_variables;
  std::vector<std::string> provenance_variables;
  RewrittenQuery():
    query(""),
    solution_variables(std::vector<std::string>()),
    provenance_variables(std::vector<std::string>()) { }
  RewrittenQuery(std::string q,
                 std::vector<std::string> sv,
                 std::vector<std::string> pv):
    query(q),
    solution_variables(sv),
    provenance_variables(pv) { }
  std::string solution_variables_str() {
    return vars_str(solution_variables);
  }
  std::string provenance_variables_str() {
    return vars_str(provenance_variables);
  }
};

/**
 * Rewrites a query to compute provenance.
 * 
 * @param query The query to rewrite.
 * @param prefix The prefix used to reify the query.
 * @param reify A function that reifies triple patterns.
 * @return The reified query, and its variables.
 */
RewrittenQuery
rewrite(std::string query,
        std::string prefix,
        std::string (*reify)(std::string, std::string));

/**
 * Rewrites a query to compute provenance.
 *
 * @param query The query to rewrite.
 * @param reify A function that reifies triple patterns.
 * @return The reified query.
 */
RewrittenQuery
rewrite_with_prefix(std::string query,
		    std::string prefix,
		    std::string (*reify)(std::string, std::string));
